#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import re

from hacking import core

RE_RELATIVE_IMPORT = re.compile(r'^from\s*[.]')


@core.flake8ext
def hacking_import_rules(logical_line, filename, noqa):
    r"""Check for imports.

    OpenStack HACKING guide recommends one import per line:
    Do not import more than one module per line

    Examples:
    Okay: from nova.compute import api
    H301: from nova.compute import api, utils


    Do not use wildcard import

    Do not make relative imports

    Examples:
    Okay: from os import path
    Okay: from os import path as p
    Okay: from os import (path as p)
    Okay: import os.path
    Okay: from nova.compute import rpcapi
    Okay: from six.moves.urllib import parse
    H303: from os.path import *
    H304: from .compute import rpcapi
    """
    # TODO(jogo): make the following doctests pass:
    #            H301: import os, sys
    # TODO(mordred: We need to split this into different checks so that they
    # can be disabled by command line switches properly

    if noqa:
        return

    split_line = logical_line.split()
    split_line_len = len(split_line)
    if (split_line_len > 1 and split_line[0] in ('import', 'from') and
            not core.is_import_exception(split_line[1])):
        pos = logical_line.find(',')
        if pos != -1:
            if split_line[0] == 'from':
                yield pos, "H301: one import per line"
        pos = logical_line.find('*')
        if pos != -1:
            yield pos, "H303: No wildcard (*) import."
            return

        if split_line_len in (2, 4, 6) and split_line[1] != "__future__":
            if 'from' == split_line[0] and split_line_len > 3:
                mod = '.'.join((split_line[1], split_line[3]))
                if core.is_import_exception(mod):
                    return
                if RE_RELATIVE_IMPORT.search(logical_line):
                    yield logical_line.find('.'), (
                        "H304: No relative imports. '%s' is a relative import"
                        % logical_line)
                    return


@core.flake8ext
def hacking_import_alphabetical(logical_line, blank_before, previous_logical,
                                indent_level, previous_indent_level):
    r"""Check for imports in alphabetical order.

    OpenStack HACKING guide recommendation for imports:
    imports in human alphabetical order

    Okay: import os\nimport sys\n\nimport nova\nfrom nova import test
    Okay: import os\nimport sys
    H306: import sys\nimport os
    Okay: import sys\n\n# foo\nimport six
    """
    # handle import x
    # use .lower since capitalization shouldn't dictate order
    if blank_before < 1 and indent_level == previous_indent_level:
        split_line = core.import_normalize(logical_line.
                                           strip()).lower().split()
        split_previous = core.import_normalize(previous_logical.
                                               strip()).lower().split()
        length = [2, 4]
        if (len(split_line) in length and len(split_previous) in length and
                split_line[0] == "import" and split_previous[0] == "import"):
            if split_line[1] < split_previous[1]:
                yield (0, "H306: imports not in alphabetical order (%s, %s)"
                       % (split_previous[1], split_line[1]))



# Copyright (c) 2012, Cloudscaling
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""OpenStack HACKING file compliance testing

Built as a sets of pycodestyle checks using flake8.
"""

import gettext
import sys

from hacking import config

# Import tests need to inject _ properly into the builtins
kwargs = {}
if sys.version_info[0] < 3:
    # In Python2, ensure that the _() that gets installed into built-ins
    # always returns unicodes. This matches the default behavior under Python
    # 3, although the keyword argument is not present in the Python 3 API.
    kwargs['unicode'] = True
gettext.install('hacking', **kwargs)


def flake8ext(f):
    f.name = __name__
    f.version = '0.0.1'
    f.skip_on_py3 = False
    if not hasattr(f, 'off_by_default'):
        f.off_by_default = False
    return f


def off_by_default(f):
    """Decorator to turn check off by default.

    To enable the check use the flake8 select setting in
    tox.ini.

    flake8 documentation:
    http://flake8.readthedocs.org/en/latest/extensions.html.
    """
    f.off_by_default = True
    return f


def skip_on_py3(f):
    f.skip_on_py3 = True
    return f

# Error code block layout

# H1xx comments
# H20x except
# H23x Python 2.x -> 3.x portability issues
# H3xx imports
# H4xx docstrings
# H5xx dictionaries/lists
# H6xx calling methods
# H7xx localization
# H8xx git commit messages
# H9xx other


CONF = config.Config('hacking')


DEFAULT_IMPORT_EXCEPTIONS = [
    'sqlalchemy',
    'migrate',
]

IMPORT_EXCEPTIONS = CONF.get_multiple('import_exceptions', default=[])
IMPORT_EXCEPTIONS += DEFAULT_IMPORT_EXCEPTIONS


def is_import_exception(mod):
    """Check module name to see if import has been whitelisted.

       Import based rules should not run on any whitelisted module
       """
    return (mod in IMPORT_EXCEPTIONS or
            any(mod.startswith(m + '.') for m in IMPORT_EXCEPTIONS))


def import_normalize(line):
    # convert "from x import y" to "import x.y"
    # handle "from x import y as z" to "import x.y as z"
    split_line = line.split()
    if ("import" in line and line.startswith("from ") and "," not in line and
            split_line[2] == "import" and split_line[3] != "*" and
            split_line[1] != "__future__" and
            (len(split_line) == 4 or
             (len(split_line) == 6 and split_line[4] == "as"))):
        return "import %s.%s" % (split_line[1], split_line[3])
    else:
        return line


class GlobalCheck(object):
    """Base class for checks that should be run only once."""

    name = None
    version = '0.0.1'
    skip_on_py3 = False
    _has_run = set()

    def __init__(self, tree, *args):
        pass

    def run(self):
        """Make run a no-op if run() has been called before.

        Store in a global registry the list of checks we've run. If we have
        run that one before, just skip doing anything the subsequent times.
        This way, since pycodestyle is file/line based, we don't wind
        up re-running a check on a git commit message over and over again.
        """
        if self.name and self.name not in self.__class__._has_run:
            self.__class__._has_run.add(self.name)
            ret = self.run_once()
            if ret is not None:
                yield ret

    def run_once(self):
        pass



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import ast
import re

from hacking import core


@core.off_by_default
@core.flake8ext
class MockAutospecCheck(object):
    """Check for 'autospec' in mock.patch/mock.patch.object calls

    Okay: mock.patch('target_module_1', autospec=True)
    Okay: mock.patch('target_module_1', autospec=False)
    Okay: mock.patch('target_module_1', autospec=None)
    Okay: mock.patch('target_module_1', defined_mock)
    Okay: mock.patch('target_module_1', new=defined_mock)
    Okay: mock.patch('target_module_1', new_callable=SomeFunc)
    Okay: mock.patch('target_module_1', defined_mock)
    Okay: mock.patch('target_module_1', spec=1000)
    Okay: mock.patch('target_module_1', spec_set=['data'])
    Okay: mock.patch('target_module_1', wraps=some_obj)

    H210: mock.patch('target_module_1')
    Okay: mock.patch('target_module_1')  # noqa
    H210: mock.patch('target_module_1', somearg=23)
    Okay: mock.patch('target_module_1', somearg=23)  # noqa

    Okay: mock.patch.object('target_module_2', 'attribute', autospec=True)
    Okay: mock.patch.object('target_module_2', 'attribute', autospec=False)
    Okay: mock.patch.object('target_module_2', 'attribute', autospec=None)
    Okay: mock.patch.object('target_module_2', 'attribute', new=defined_mock)
    Okay: mock.patch.object('target_module_2', 'attribute', defined_mock)
    Okay: mock.patch.object('target_module_2', 'attribute', new_callable=AFunc)
    Okay: mock.patch.object('target_module_2', 'attribute', spec=3)
    Okay: mock.patch.object('target_module_2', 'attribute', spec_set=[3])
    Okay: mock.patch.object('target_module_2', 'attribute', wraps=some_obj)


    H210: mock.patch.object('target_module_2', 'attribute', somearg=2)
    H210: mock.patch.object('target_module_2', 'attribute')

    """

    name = "mock_check"
    version = "1.00"

    def __init__(self, tree, filename):
        self.filename = filename
        self.tree = tree

    def run(self):
        mcv = MockCheckVisitor(self.filename)
        mcv.visit(self.tree)
        for message in mcv.messages:
            yield message


class MockCheckVisitor(ast.NodeVisitor):
    # Patchers we are looking for and minimum number of 'args' without
    # 'autospec' to not be flagged
    patchers = {'mock.patch': 2, 'mock.patch.object': 3}
    spec_keywords = {"autospec", "new", "new_callable", "spec", "spec_set",
                     "wraps"}

    def __init__(self, filename):
        super(MockCheckVisitor, self).__init__()
        self.messages = []
        self.filename = filename

    def check_missing_autospec(self, call_node):

        def find_autospec_keyword(keyword_node):
            for keyword_obj in keyword_node:
                keyword = keyword_obj.arg
                # If they have defined autospec or new then it is okay
                if keyword in self.spec_keywords:
                    return True
            return False

        if isinstance(call_node, ast.Call):
            func_info = FunctionNameFinder(self.filename)
            func_info.visit(call_node)

            # We are only looking at our patchers
            if func_info.function_name not in self.patchers:
                return

            min_args = self.patchers[func_info.function_name]

            if not find_autospec_keyword(call_node.keywords):
                if len(call_node.args) < min_args:
                    self.messages.append(
                        (call_node.lineno, call_node.col_offset,
                         "H210 Missing 'autospec' or 'spec_set' keyword in "
                         "mock.patch/mock.patch.object", MockCheckVisitor)
                    )

    def visit_Call(self, node):
        self.check_missing_autospec(node)
        self.generic_visit(node)


class FunctionNameFinder(ast.NodeVisitor):
    """Finds the name of the function"""
    def __init__(self, filename):
        super(FunctionNameFinder, self).__init__()
        self._func_name = []
        self.filename = filename

    @property
    def function_name(self):
        return '.'.join(reversed(self._func_name))

    def visit_Name(self, node):
        self._func_name.append(node.id)
        self.generic_visit(node)

    def visit_Attribute(self, node):
        try:
            self._func_name.append(node.attr)
            self._func_name.append(node.value.id)
        except AttributeError:
            self.generic_visit(node)

    def visit(self, node):
        # If we get called with an ast.Call node, then work on the 'node.func',
        # as we want the function name.
        if isinstance(node, ast.Call):
            return super(FunctionNameFinder, self).visit(node.func)
        return super(FunctionNameFinder, self).visit(node)


third_party_mock = re.compile('^import.mock')
from_third_party_mock = re.compile('^from.mock.import')


@core.flake8ext
def hacking_no_third_party_mock(logical_line, noqa):
    """Check for use of mock instead of unittest.mock.

    Projects have had issues with using mock without including it in their
    requirements, thinking it is the standard library version. This makes it so
    these projects need to explicitly turn off this check to make it clear they
    intended to use it.

    Okay: from unittest import mock
    Okay: from unittest.mock import patch
    Okay: import unittest.mock
    H216: import mock
    H216: from mock import patch
    Okay: try: import mock
    Okay: import mock  # noqa
    """
    msg = ('H216: The unittest.mock module should be used rather than the '
           'third party mock package unless actually needed. If so, disable '
           'the H216 check in hacking config and ensure mock is declared in '
           "the project's requirements.")

    if noqa:
        return

    if (re.match(third_party_mock, logical_line) or
            re.match(from_third_party_mock, logical_line)):
        yield (0, msg)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

# module cannot be called except since that is a reserved word

import ast
import re

from hacking import core

RE_ASSERT_RAISES_EXCEPTION = re.compile(r"self\.assertRaises\(Exception[,\)]")
RE_ASSERT_TRUE_INST = re.compile(
    r"(.)*assertTrue\(isinstance\((\w|\.|\'|\"|\[|\])+, "
    r"(\w|\.|\'|\"|\[|\])+\)\)")
RE_ASSERT_EQUAL_TYPE = re.compile(
    r"(.)*assertEqual\(type\((\w|\.|\'|\"|\[|\])+\), "
    r"(\w|\.|\'|\"|\[|\])+\)")
RE_ASSERT_EQUAL_IN_START_WITH_TRUE_OR_FALSE = re.compile(
    r"assertEqual\("
    r"(True|False), (\w|[][.'\"])+ in (\w|[][.'\", ])+\)")
RE_ASSERT_RAISES_REGEXP = re.compile(r"assertRaisesRegexp\(")
# NOTE(snikitin): Next two regexes weren't united to one for more readability.
#                 asse_true_false_with_in_or_not_in regex checks
#                 assertTrue/False(A in B) cases where B argument has no spaces
#                 asse_true_false_with_in_or_not_in_spaces regex checks cases
#                 where B argument has spaces and starts/ends with [, ', ".
#                 For example: [1, 2, 3], "some string", 'another string'.
#                 We have to separate these regexes to escape a false positives
#                 results. B argument should have spaces only if it starts
#                 with [, ", '. Otherwise checking of string
#                 "assertFalse(A in B and C in D)" will be false positives.
#                 In this case B argument is "B and C in D".
RE_ASSERT_TRUE_FALSE_WITH_IN_OR_NOT_IN = re.compile(
    r"assert(True|False)\("
    r"(\w|[][.'\"])+( not)? in (\w|[][.'\",])+(, .*)?\)")
RE_ASSERT_TRUE_FALSE_WITH_IN_OR_NOT_IN_SPACES = re.compile(
    r"assert(True|False)"
    r"\((\w|[][.'\"])+( not)? in [\[|'|\"](\w|[][.'\", ])+"
    r"[\[|'|\"](, .*)?\)")
RE_ASSERT_EQUAL_IN_END_WITH_TRUE_OR_FALSE = re.compile(
    r"assertEqual\("
    r"(\w|[][.'\"])+ in (\w|[][.'\", ])+, (True|False)\)")


@core.flake8ext
def hacking_except_format(logical_line, noqa):
    r"""Check for 'except:'.

    OpenStack HACKING guide recommends not using except:
    Do not write "except:", use "except Exception:" at the very least

    Okay: try:\n    pass\nexcept Exception:\n    pass
    H201: try:\n    pass\nexcept:\n    pass
    Okay: try:\n    pass\nexcept:  # noqa\n    pass
    """
    if noqa:
        return
    if logical_line.startswith("except:"):
        yield 6, "H201: no 'except:' at least use 'except Exception:'"


@core.flake8ext
def hacking_except_format_assert(logical_line, noqa):
    r"""Check for 'assertRaises(Exception'.

    OpenStack HACKING guide recommends not using assertRaises(Exception...):
    Do not use overly broad Exception type

    Okay: self.assertRaises(NovaException, foo)
    Okay: self.assertRaises(ExceptionStrangeNotation, foo)
    H202: self.assertRaises(Exception, foo)
    H202: self.assertRaises(Exception)
    Okay: self.assertRaises(Exception)  # noqa
    Okay: self.assertRaises(Exception, foo)  # noqa
    """
    if noqa:
        return
    if RE_ASSERT_RAISES_EXCEPTION.search(logical_line):
        yield 1, "H202: assertRaises Exception too broad"


def is_none(node):
    '''Check whether an AST node corresponds to None.'''
    return isinstance(node, ast.Constant) and node.value is None


def _get_local_func_name(node):
    if isinstance(node.func, ast.Attribute):
        return node.func.attr
    elif isinstance(node.func, ast.Name):
        return node.func.id
    else:
        return None


class NoneArgChecker(ast.NodeVisitor):
    '''NodeVisitor to check function calls for None arguments.

    :param func_name: only check calls to functions with this name
    :param num_args:  number of arguments to check for None

    self.none_found will be True if any None arguments were found.
    '''

    def __init__(self, func_name, num_args=2):
        self.func_name = func_name
        self.num_args = num_args
        self.none_found = False

    def visit_Call(self, node):
        local_func_name = _get_local_func_name(node)

        if local_func_name == self.func_name:
            args_to_check = node.args[:self.num_args]
            self.none_found |= any(is_none(x) for x in args_to_check)
        self.generic_visit(node)


@core.flake8ext
@core.off_by_default
def hacking_assert_is_none(logical_line, noqa):
    """Use assertIs(Not)None to check for None in assertions.

    Okay: self.assertEqual('foo', 'bar')
    Okay: self.assertNotEqual('foo', {}.get('bar', None))
    Okay: self.assertIs('foo', 'bar')
    Okay: self.assertIsNot('foo', 'bar', None)
    Okay: foo(self.assertIsNot('foo', 'bar'))
    H203: self.assertEqual(None, 'foo')
    H203: self.assertNotEqual('foo', None)
    H203: self.assertIs(None, 'foo', 'bar')
    H203: self.assertIsNot('foo', None, 'bar')
    H203: foo(self.assertIsNot('foo', None, 'bar'))
    Okay: self.assertEqual(None, 'foo')  # noqa
    Okay: self.assertIs(None, 'foo')  # noqa
    Okay: self.assertIsNone('foo')
    """
    if noqa:
        return
    for func_name in ('assertEqual', 'assertIs', 'assertNotEqual',
                      'assertIsNot'):
        try:
            start = logical_line.index('.%s(' % func_name) + 1
        except ValueError:
            continue
        checker = NoneArgChecker(func_name)
        try:
            parsed_logical_line = ast.parse(logical_line)
        except SyntaxError:
            # let flake8 catch this itself
            # https://github.com/PyCQA/flake8/issues/1948
            continue
        checker.visit(parsed_logical_line)
        if checker.none_found:
            yield start, "H203: Use assertIs(Not)None to check for None"


class AssertTrueFalseChecker(ast.NodeVisitor):
    '''NodeVisitor to find "assert[True|False](some comparison)" statements.

    :param method_names: methods to look for: assertTrue and/or assertFalse
    :param ops: list of comparisons we want to look for (objects from the ast
                module)
    '''

    def __init__(self, method_names, ops):
        self.method_names = method_names
        self.ops = tuple(ops)
        self.error = False

    def visit_Call(self, node):
        # No need to keep visiting the AST if we already found something.
        if self.error:
            return

        self.generic_visit(node)

        local_func_name = _get_local_func_name(node)

        if (local_func_name in self.method_names and
                len(node.args) == 1 and
                isinstance(node.args[0], ast.Compare) and
                len(node.args[0].ops) == 1 and
                isinstance(node.args[0].ops[0], self.ops)):
            self.error = True


@core.flake8ext
@core.off_by_default
def hacking_assert_equal(logical_line, noqa):
    r"""Check that self.assertEqual and self.assertNotEqual are used.

    Okay: self.assertEqual(x, y)
    Okay: self.assertNotEqual(x, y)
    H204: self.assertTrue(x == y)
    H204: self.assertTrue(x != y)
    H204: self.assertFalse(x == y)
    H204: self.assertFalse(x != y)
    """
    if noqa:
        return

    methods = ['assertTrue', 'assertFalse']
    for method in methods:
        start = logical_line.find('.%s' % method) + 1
        if start != 0:
            break
    else:
        return
    comparisons = [ast.Eq, ast.NotEq]
    checker = AssertTrueFalseChecker(methods, comparisons)
    try:
        parsed_logical_line = ast.parse(logical_line)
    except SyntaxError:
        # let flake8 catch this itself
        # https://github.com/PyCQA/flake8/issues/1948
        return
    checker.visit(parsed_logical_line)
    if checker.error:
        yield start, 'H204: Use assert(Not)Equal()'


@core.flake8ext
@core.off_by_default
def hacking_assert_greater_less(logical_line, noqa):
    r"""Check that self.assert{Greater,Less}[Equal] are used.

    Okay: self.assertGreater(x, y)
    Okay: self.assertGreaterEqual(x, y)
    Okay: self.assertLess(x, y)
    Okay: self.assertLessEqual(x, y)
    H205: self.assertTrue(x > y)
    H205: self.assertTrue(x >= y)
    H205: self.assertTrue(x < y)
    H205: self.assertTrue(x <= y)
    """
    if noqa:
        return

    methods = ['assertTrue', 'assertFalse']
    for method in methods:
        start = logical_line.find('.%s' % method) + 1
        if start != 0:
            break
    else:
        return
    comparisons = [ast.Gt, ast.GtE, ast.Lt, ast.LtE]
    checker = AssertTrueFalseChecker(methods, comparisons)
    try:
        parsed_logical_line = ast.parse(logical_line)
    except SyntaxError:
        # let flake8 catch this itself
        # https://github.com/PyCQA/flake8/issues/1948
        return
    checker.visit(parsed_logical_line)
    if checker.error:
        yield start, 'H205: Use assert{Greater,Less}[Equal]'


@core.flake8ext
def hacking_assert_true_instance(logical_line):
    """Check for assertTrue(isinstance(a, b)) sentences

    H211
    """
    if RE_ASSERT_TRUE_INST.match(logical_line):
        yield (
            0,
            "H211: Use assert{Is,IsNot}instance")


@core.flake8ext
def hacking_assert_equal_type(logical_line):
    """Check for assertEqual(type(A), B) sentences

    H212
    """
    if RE_ASSERT_EQUAL_TYPE.match(logical_line):
        yield (
            0,
            "H211: Use assert{Is,IsNot}instance")


@core.flake8ext
def hacking_assert_raises_regexp(logical_line):
    """Check for usage of deprecated assertRaisesRegexp

    H213
    """
    res = RE_ASSERT_RAISES_REGEXP.search(logical_line)
    if res:
        yield (
            0,
            "H213: assertRaisesRegex must be used instead "
            "of assertRaisesRegexp")


@core.flake8ext
def hacking_assert_true_or_false_with_in(logical_line):
    """Check for assertTrue/False(A in B), assertTrue/False(A not in B),

    assertTrue/False(A in B, message) or assertTrue/False(A not in B, message)
    sentences.
    H214
    """
    res = (RE_ASSERT_TRUE_FALSE_WITH_IN_OR_NOT_IN.search(logical_line)
           or RE_ASSERT_TRUE_FALSE_WITH_IN_OR_NOT_IN_SPACES.search
           (logical_line))
    if res:
        yield (
            0,
            "H214: Use assertIn/NotIn(A, B) rather than "
            "assertTrue/False(A in/not in B) when checking collection "
            "contents.")


@core.flake8ext
def hacking_assert_equal_in(logical_line):
    """Check for assertEqual(A in B, True), assertEqual(True, A in B),

    assertEqual(A in B, False) or assertEqual(False, A in B) sentences

    H215
    """
    res = (RE_ASSERT_EQUAL_IN_START_WITH_TRUE_OR_FALSE.search(logical_line) or
           RE_ASSERT_EQUAL_IN_END_WITH_TRUE_OR_FALSE.search(logical_line))
    if res:
        yield (
            0,
            "H215: Use assertIn/NotIn(A, B) rather than "
            "assertEqual(A in B, True/False) when checking collection "
            "contents.")



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import re
import tokenize

from hacking import core


FORMAT_RE = re.compile(r"%(?:"
                       r"%|"           # Ignore plain percents
                       r"(\(\w+\))?"   # mapping key
                       r"([#0 +-]?"    # flag
                       r"(?:\d+|\*)?"  # width
                       r"(?:\.\d+)?"   # precision
                       r"[hlL]?"       # length mod
                       r"\w))")        # type


class LocalizationError(Exception):
    pass


def check_i18n():
    """Generator that checks token stream for localization errors.

    Expects tokens to be ``send``ed one by one.
    Raises LocalizationError if some error is found.
    """
    while True:
        try:
            token_type, text, _, _, line = yield
        except GeneratorExit:
            return

        if text == "def" and token_type == tokenize.NAME:
            # explicitly ignore function definitions, as oslo defines these
            return
        if (token_type == tokenize.NAME and
                text in ["_", "_LI", "_LW", "_LE", "_LC"]):

            while True:
                token_type, text, start, _, _ = yield
                if token_type != tokenize.NL:
                    break
            if token_type != tokenize.OP or text != "(":
                continue  # not a localization call

            format_string = ''
            while True:
                token_type, text, start, _, _ = yield
                if token_type == tokenize.STRING:
                    format_string += eval(text)
                elif token_type == tokenize.NL:
                    pass
                else:
                    break

            if not format_string:
                raise LocalizationError(
                    start, "H701: Empty localization string")
            if token_type != tokenize.OP:
                raise LocalizationError(
                    start, "H701: Invalid localization call")
            if text != ")":
                if text == "%":
                    raise LocalizationError(
                        start,
                        "H702: Formatting operation should be outside"
                        " of localization method call")
                elif text == "+":
                    raise LocalizationError(
                        start,
                        "H702: Use bare string concatenation instead of +")
                else:
                    raise LocalizationError(
                        start, "H702: Argument to _, _LI, _LW, _LC, or _LE "
                        "must be just a string")

            format_specs = FORMAT_RE.findall(format_string)
            positional_specs = [(key, spec) for key, spec in format_specs
                                if not key and spec]
            # not spec means %%, key means %(smth)s
            if len(positional_specs) > 1:
                raise LocalizationError(
                    start, "H703: Multiple positional placeholders")


@core.flake8ext
def hacking_localization_strings(logical_line, tokens, noqa):
    r"""Check localization in line.

    Okay: _("This is fine")
    Okay: _LI("This is fine")
    Okay: _LW("This is fine")
    Okay: _LE("This is fine")
    Okay: _LC("This is fine")
    Okay: _("This is also fine %s")
    Okay: _("So is this %s, %(foo)s") % {foo: 'foo'}
    H701: _('')
    Okay: def _(msg):\n    pass
    Okay: def _LE(msg):\n    pass
    H701: _LI('')
    H701: _LW('')
    H701: _LE('')
    H701: _LC('')
    Okay: _('')  # noqa
    H702: _("Bob" + " foo")
    H702: _LI("Bob" + " foo")
    H702: _LW("Bob" + " foo")
    H702: _LE("Bob" + " foo")
    H702: _LC("Bob" + " foo")
    Okay: _("Bob" + " foo")  # noqa
    H702: _("Bob %s" % foo)
    H702: _LI("Bob %s" % foo)
    H702: _LW("Bob %s" % foo)
    H702: _LE("Bob %s" % foo)
    H702: _LC("Bob %s" % foo)
    H702: _("%s %s" % (foo, bar))
    H703: _("%s %s") % (foo, bar)
    """
    if noqa:
        return
    gen = check_i18n()
    next(gen)
    try:
        list(map(gen.send, tokens))
        gen.close()
    except LocalizationError as e:
        yield e.args

# TODO(jogo) Dict and list objects



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import tokenize

from hacking import core

START_DOCSTRING_TRIPLE = ['u"""', 'r"""', '"""', "u'''", "r'''", "'''"]
END_DOCSTRING_TRIPLE = ['"""', "'''"]


@core.flake8ext
def hacking_docstring_start_space(physical_line, previous_logical, tokens):
    r"""Check for docstring not starting with space.

    OpenStack HACKING guide recommendation for docstring:
    Docstring should not start with space

    Okay: def foo():\n    '''This is good.'''
    Okay: def foo():\n    r'''This is good.'''
    Okay: def foo():\n    a = ''' This is not a docstring.'''
    Okay: def foo():\n    pass\n    ''' This is not.'''
    H401: def foo():\n    ''' This is not.'''
    H401: def foo():\n    r''' This is not.'''
    """
    docstring = is_docstring(tokens, previous_logical)
    if docstring:
        start, start_triple = _find_first_of(docstring, START_DOCSTRING_TRIPLE)
        if docstring[len(start_triple)] == ' ':
            # docstrings get tokenized on the last line of the docstring, so
            # we don't know the exact position.
            return (0, "H401: docstring should not start with"
                    " a space")


@core.flake8ext
def hacking_docstring_multiline_end(physical_line, previous_logical, tokens):
    r"""Check multi line docstring end.

    OpenStack HACKING guide recommendation for docstring:
    Docstring should end on a new line

    Okay: '''foobar\nfoo\nbar\n'''
    Okay: def foo():\n    '''foobar\n\nfoo\nbar\n'''
    Okay: class Foo(object):\n    '''foobar\n\nfoo\nbar\n'''
    Okay: def foo():\n    a = '''not\na\ndocstring'''
    Okay: def foo():\n    a = '''not\na\ndocstring'''  # blah
    Okay: def foo():\n    pass\n'''foobar\nfoo\nbar\n   d'''
    H403: def foo():\n    '''foobar\nfoo\nbar\ndocstring'''
    H403: def foo():\n    '''foobar\nfoo\nbar\npretend raw: r'''
    H403: class Foo(object):\n    '''foobar\nfoo\nbar\ndocstring'''\n\n
    """
    docstring = is_docstring(tokens, previous_logical)
    if docstring:
        if '\n' not in docstring:
            # not a multi line
            return
        else:
            last_line = docstring.split('\n')[-1]
        pos = max(last_line.rfind(i) for i in END_DOCSTRING_TRIPLE)
        if len(last_line[:pos].strip()) > 0:
            # Something before the end docstring triple
            return (pos,
                    "H403: multi line docstrings should end on a new line")


@core.flake8ext
def hacking_docstring_multiline_start(physical_line, previous_logical, tokens):
    r"""Check multi line docstring starts immediately with summary.

    OpenStack HACKING guide recommendation for docstring:
    Docstring should start with a one-line summary, less than 80 characters.

    Okay: '''foobar\n\nfoo\nbar\n'''
    Okay: def foo():\n    a = '''\nnot\na docstring\n'''
    H404: def foo():\n    '''\nfoo\nbar\n'''\n\n
    H404: def foo():\n    r'''\nfoo\nbar\n'''\n\n
    """
    docstring = is_docstring(tokens, previous_logical)
    if docstring:
        if '\n' not in docstring:
            # single line docstring
            return
        start, start_triple = _find_first_of(docstring, START_DOCSTRING_TRIPLE)
        lines = docstring.split('\n')
        if lines[0].strip() == start_triple:
            # docstrings get tokenized on the last line of the docstring, so
            # we don't know the exact position.
            return (0, "H404: multi line docstring "
                    "should start without a leading new line")


@core.flake8ext
def hacking_docstring_summary(physical_line, previous_logical, tokens):
    r"""Check multi line docstring summary is separated with empty line.

    OpenStack HACKING guide recommendation for docstring:
    Docstring should start with a one-line summary, less than 80 characters.

    Okay: def foo():\n    a = '''\nnot\na docstring\n'''
    Okay: '''foobar\n\nfoo\nbar\n'''
    H405: def foo():\n    '''foobar\nfoo\nbar\n'''
    H405: def foo():\n    r'''foobar\nfoo\nbar\n'''
    H405: def foo():\n    '''foobar\n'''
    """
    docstring = is_docstring(tokens, previous_logical)
    if docstring:
        if '\n' not in docstring:
            # not a multi line docstring
            return
        lines = docstring.split('\n')
        if len(lines) > 1 and len(lines[1].strip()) != 0:
            # docstrings get tokenized on the last line of the docstring, so
            # we don't know the exact position.
            return (0, "H405: multi line docstring "
                    "summary not separated with an empty line")


def is_docstring(tokens, previous_logical):
    """Return found docstring

    'A docstring is a string literal that occurs as the first statement in a
    module, function, class,'
    http://www.python.org/dev/peps/pep-0257/#what-is-a-docstring
    """
    for token_type, text, start, _, _ in tokens:
        if token_type == tokenize.STRING:
            break
        elif token_type != tokenize.INDENT:
            return False
    else:
        return False
    line = text.lstrip()
    start, start_triple = _find_first_of(line, START_DOCSTRING_TRIPLE)
    if (previous_logical.startswith("def ") or
            previous_logical.startswith("class ")):
        if start == 0:
            return text


def _find_first_of(line, substrings):
    """Find earliest occurrence of one of substrings in line.

    Returns pair of index and found substring, or (-1, None)
    if no occurrences of any of substrings were found in line.
    """
    starts = ((line.find(i), i) for i in substrings)
    found = [(i, sub) for i, sub in starts if i != -1]
    if found:
        return min(found)
    else:
        return -1, None



# Copyright (c) 2013 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hacking.checks import comments
from hacking import tests


class CoreTestCase(tests.TestCase):
    def test_H102_none(self):
        """Verify that the H102 check finds an SPDX header"""
        self.assertEqual(
            (0, 'H102: Apache 2.0 license header not found'),
            comments.hacking_has_license(
                None,
                None,
                [
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                ],
                1,
            ),
        )

    def test_H102_full(self):
        """Verify that the H102 check finds an SPDX header"""
        self.assertIsNone(comments.hacking_has_license(
            None,
            None,
            [
                '# foo',
                '# Licensed under the Apache License, Version 2.0',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
            ],
            1,
        ))

    def test_H102_SPDX(self):
        """Verify that the H102 check finds an SPDX header"""
        self.assertIsNone(comments.hacking_has_license(
            None,
            None,
            [
                '# foo',
                '# SPDX-License-Identifier: Apache-2.0',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
            ],
            1,
        ))

    def test_H103_full_fail(self):
        """Verify that the H103 check finds an SPDX header"""
        self.assertEqual(
            (2, 'H103: Header does not match Apache 2.0 License notice\n'
                '<license>!=<apache2>:\n'
                '\'Licensed under the Apache License, Version 2.0 foo bar foo '
                'bar foo bar foo bar foo bar\' !=\n'
                '\'Licensed under the Apache License, Version 2.0 (the '
                '"License"); you may not use this file except in compliance '
                'with the License. You may obtain a copy of the License at '
                'http://www.apache.org/licenses/LICENSE-2.0 Unless required '
                'by applicable law or agreed to in writing, software '
                'distributed under the License is distributed on an "AS IS" '
                'BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either '
                'express or implied. See the License for the specific '
                'language governing permissions and limitations under the '
                'License.\''),
            comments.hacking_has_correct_license(
                None,
                None,
                [
                    '# foo',
                    '# Licensed under the Apache License, Version 2.0',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                    '# foo',
                    '# bar',
                ],
                1,
            ),
        )

    def test_H103_full(self):
        """Verify that the H103 check finds an SPDX header"""
        self.assertIsNone(comments.hacking_has_correct_license(
            None,
            None,
            [
                """
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
                """  # noqa
            ],
            1,
        ))

    def test_H103_SPDX(self):
        """Verify that the H103 check finds an SPDX header"""
        self.assertIsNone(comments.hacking_has_correct_license(
            None,
            None,
            [
                '# foo',
                '# SPDX-License-Identifier: Apache-2.0',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
                '# foo',
                '# bar',
            ],
            1,
        ))

    def test_H104_regex(self):
        """Verify that the H104 regex matches correct lines."""
        self.assertTrue(comments.hacking_has_only_comments(
            None,
            None,
            ['# foo',
             '# bar'],
            1))
        self.assertTrue(comments.hacking_has_only_comments(
            None,
            None,
            ['# foo',
             '# bar',
             ''],
            1))
        self.assertTrue(comments.hacking_has_only_comments(
            None,
            None,
            ['# foo',
             '   ',
             '# bar'],
            1))

        self.assertIsNone(comments.hacking_has_only_comments(
            None,
            None,
            ['# foo',
             '   ',
             '"""foobar"""'],
            1))
        self.assertIsNone(comments.hacking_has_only_comments(
            None,
            None,
            ['# foo',
             '',
             'print(42)'],
            1))
        self.assertIsNone(comments.hacking_has_only_comments(
            None,
            None,
            ['# foo'],
            100))

    def test_H105(self):
        self.assertTrue(comments.hacking_no_author_tags(
            '# @author: Foo Bar'))

        self.assertTrue(comments.hacking_no_author_tags(
            '# @Author: Foo Bar'))

        self.assertTrue(comments.hacking_no_author_tags(
            '# author: Foo Bar'))

        self.assertTrue(comments.hacking_no_author_tags(
            '# authors: Foo Bar'))

        self.assertTrue(comments.hacking_no_author_tags(
            '# Author: Foo Bar'))

        self.assertTrue(comments.hacking_no_author_tags(
            '# Authors: Foo Bar'))

        self.assertTrue(comments.hacking_no_author_tags(
            '.. moduleauthor:: Foo Bar'))



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import re

from hacking import core


log_string = re.compile(r".*LOG\.(?:error|warn|warning|info"
                        r"|critical|exception|debug)")


@core.flake8ext
def hacking_no_cr(physical_line):
    r"""Check that we only use newlines not carriage returns.

    Okay: import os\nimport sys
    # pep8 doesn't yet replace \r in strings, will work on an
    # upstream fix
    H903 import os\r\nimport sys
    """
    if '\r' in physical_line:
        yield (0, "H903: Windows style line endings not allowed in code")


@core.flake8ext
@core.off_by_default
def hacking_delayed_string_interpolation(logical_line, noqa):
    r"""String interpolation should be delayed at logging calls.

    H904: LOG.debug('Example: %s' % 'bad')
    Okay: LOG.debug('Example: %s', 'good')
    """
    msg = ("H904: String interpolation should be delayed to be "
           "handled by the logging code, rather than being done "
           "at the point of the logging call. "
           "Use ',' instead of '%'.")

    if noqa:
        return

    if log_string.match(logical_line):
        # Line is a log statement, strip out strings and see if % is used,
        # just to make sure we don't match on a format specifier in a string.
        line = re.sub(r"[\"'].+?[\"']", '', logical_line)
        # There are some cases where string formatting of the arguments are
        # needed, so don't include those when checking.
        line = re.sub(r",.*", '', line)
        if '%' in line or '.format(' in line:
            yield 0, msg



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import subprocess
import sys
import tempfile

import pkg_resources
import testscenarios
from testtools import content

import hacking.tests

SELFTEST_REGEX = re.compile(r'\b(Okay|[HEW]\d{3}):\s(.*)')
# Each scenario is (name, {lines=.., raw=..., code=...})
file_cases = []


class HackingTestCase(hacking.tests.TestCase):

    scenarios = file_cases

    def test_flake8(self):

        with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
            f.write(''.join(self.lines))

        cmd = [sys.executable, '-mflake8', '--isolated',
               '--select=%s' % self.code,
               '--enable-extensions=%s' % self.code, '--ignore=F',
               '--format=%(code)s\t%(path)s\t%(row)d', f.name]
        out, _ = subprocess.Popen(cmd, stdout=subprocess.PIPE).communicate()

        out = out.decode('utf-8')

        if self.code == 'Okay':
            self.assertEqual('', out)
        else:
            self.addDetail('reason',
                           content.text_content("Failed to trigger rule %s" %
                                                self.code))

            self.assertNotEqual('', out)
            self.assertEqual(self.code, out.split('\t')[0].rstrip(':'), out)

        os.remove(f.name)


def _get_lines(check):
    for line in check.__doc__.splitlines():
        line = line.lstrip()
        match = SELFTEST_REGEX.match(line)
        if match is None:
            continue
        yield line, match.groups()


def load_tests(loader, tests, pattern):

    for entry in pkg_resources.iter_entry_points('flake8.extension'):
        if not entry.module_name.startswith('hacking.'):
            continue

        check = entry.load()
        if check.skip_on_py3:
            continue

        name = entry.attrs[0]
        for lineno, (raw, (code, source)) in enumerate(_get_lines(check)):
            lines = [part.replace(r'\t', '\t') + '\n'
                     for part in source.split(r'\n')]
            file_cases.append((
                '%s-%s-line-%s' % (entry.name, name, lineno),
                {'lines': lines, 'raw': raw, 'code': code},
            ))

    return testscenarios.load_tests_apply_scenarios(loader, tests, pattern)



# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import fixtures
import testtools


_TRUE_VALUES = ('True', 'true', '1', 'yes')


class TestCase(testtools.TestCase):
    """Test case base class for all unit tests."""

    def setUp(self):
        """Run before each test method to initialize test environment."""
        super(TestCase, self).setUp()
        test_timeout = os.environ.get('OS_TEST_TIMEOUT', 0)
        try:
            test_timeout = int(test_timeout)
        except ValueError:
            # If timeout value is invalid do not set a timeout.
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

        if os.environ.get('OS_STDOUT_CAPTURE') in _TRUE_VALUES:
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if os.environ.get('OS_STDERR_CAPTURE') in _TRUE_VALUES:
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

    def assertCheckFails(self, check_func, *args, **kwargs):
        if not list(check_func(*args, **kwargs)):
            raise AssertionError("Check %s did not fail." %
                                 check_func.__name__)

    def assertCheckPasses(self, check_func, *args, **kwargs):
        try:
            self.assertCheckFails(check_func, *args, **kwargs)
        except AssertionError:
            return
        else:
            raise AssertionError("Check %s failed." % check_func.__name__)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import tokenize

from hacking import core


LOCALS_TEXT_MAP = {
    'locals': 'locals()',
    'self': 'self.__dict__'
}


@core.flake8ext
def hacking_no_locals(logical_line, tokens, noqa):
    """Do not use locals() or self.__dict__ for string formatting.

    Okay: 'locals()'
    Okay: 'locals'
    Okay: locals()
    Okay: print(locals())
    H501: print("%(something)" % locals())
    H501: LOG.info(_("%(something)") % self.__dict__)
    Okay: print("%(something)" % locals())  # noqa
    """
    if noqa:
        return
    for_formatting = False
    for token_type, text, start, _, _ in tokens:
        if text == "%" and token_type == tokenize.OP:
            for_formatting = True
        if for_formatting and token_type == tokenize.NAME:
            for k, v in LOCALS_TEXT_MAP.items():
                if text == k and v in logical_line:
                    yield (start[1],
                           "H501: Do not use %s for string formatting" % v)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from hacking import core


vim_header_re = re.compile(r"^#\s+vim?:.+")


@core.flake8ext
@core.off_by_default
def no_vim_headers(physical_line, line_number, lines):
    r"""Check for vim editor configuration in source files.

    By default vim modelines can only appear in the first or
    last 5 lines of a source file.

    Examples:
    H106: # vim: set tabstop=4 shiftwidth=4\n#\n#\n#\n#\n#
    H106: # Lic\n# vim: set tabstop=4 shiftwidth=4\n#\n#\n#\n#\n#
    H106: # Lic\n#\n#\n#\n#\n#\n#\n#\n#\n# vim: set tabstop=4 shiftwidth=4
    Okay: # Lic\n#\n#\n#\n#\n#\n#\n#
    Okay: # viminal hill is located in Rome
    Okay: # vim, ze nemluvis cesky
    """
    if ((line_number <= 5 or line_number > len(lines) - 5) and
            vim_header_re.match(physical_line)):
        return 0, "H106: Don't put vim configuration in source files"



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import textwrap
from unittest import mock

import pycodestyle

from hacking.checks import except_checks
from hacking import tests


class ExceptChecksTest(tests.TestCase):

    def _assert_has_errors(self, code, checker, expected_errors=None,
                           filename=None):
        actual_errors = [e[:3] for e in
                         self._run_check(code, checker, filename)]
        self.assertEqual(expected_errors or [], actual_errors)

    # We are patching pycodestyle so that only the check under test is actually
    # installed.
    @mock.patch('pycodestyle._checks',
                {'physical_line': {}, 'logical_line': {}, 'tree': {}})
    def _run_check(self, code, checker, filename=None):
        pycodestyle.register_check(checker)

        lines = textwrap.dedent(code).lstrip().splitlines(True)

        checker = pycodestyle.Checker(filename=filename, lines=lines)
        # NOTE(sdague): the standard reporter has printing to stdout
        # as a normal part of check_all, which bleeds through to the
        # test output stream in an unhelpful way. This blocks that printing.
        with mock.patch('pycodestyle.StandardReport.get_file_results'):
            checker.check_all()
        checker.report._deferred_print.sort()
        return checker.report._deferred_print

    def test_hacking_assert_true_instance(self):
        self.assertEqual(
            len(list(except_checks.hacking_assert_true_instance(
                "self.assertTrue(isinstance(e, "
                "exception.BuildAbortException))"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_instance(
                "self.assertTrue()"
            ))),
            0)

    def test_hacking_assert_equal_in(self):
        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual(a in b, True)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual('str' in 'string', True)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual(any(a==1 for a in b), True)"
            ))),
            0)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual(a in b, True)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual('str' in 'string', True)"
            ))),
            1)

        self.assertEqual(len(list(except_checks.hacking_assert_equal_in(
            "self.assertEqual(any(a==1 for a in b, True))"
        ))),
            0)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual(a in b, False)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual('str' in 'string', False)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual(any(a==1 for a in b), False)"
            ))),
            0)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual(a in b, False)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual('str' in 'string', False)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_in(
                "self.assertEqual(any(a==1 for a in b, False))"
            ))),
            0)

    def test_hacking_assert_true_or_false_with_in_or_not_in(self):
        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(A in B)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertFalse(A in B)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(A not in B)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertFalse(A not in B)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(A in B, 'some message')"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertFalse(A in B, 'some message')"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(A not in B, 'some message')"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertFalse(A not in B, 'some message')"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(A in 'some string with spaces')"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(A in ['1', '2', '3'])"
            ))),
            1)

        self.assertEqual(len(
            list(except_checks.hacking_assert_true_or_false_with_in(
                 "self.assertTrue(A in [1, 2, 3])"
                 ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(any(A > 5 for A in B))"
            ))),
            0)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertTrue(any(A > 5 for A in B), 'some message')"
            ))),
            0)

        self.assertEqual(
            len(list(except_checks.hacking_assert_true_or_false_with_in(
                "self.assertFalse(some in list1 and some2 in list2)"
            ))),
            0)

    def test_hacking_oslo_assert_raises_regexp(self):
        code = """
               self.assertRaisesRegexp(ValueError,
                                       "invalid literal for.*XYZ'$",
                                       int,
                                       'XYZ')
               """
        self._assert_has_errors(
            code,
            except_checks.hacking_assert_raises_regexp, expected_errors=[
                (1, 0, "H213")
            ])

    def test_hacking_assert_equal_type(self):
        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_type(
                "self.assertEqual(type(als['QuicAssist']), list)"
            ))),
            1)

        self.assertEqual(
            len(list(except_checks.hacking_assert_equal_type(
                "self.assertTrue()"
            ))),
            0)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import fixtures

from hacking import config
from hacking import core
from hacking import tests


TEST_TOX_INI = """[hacking]
import_exceptions =
    a.b.c
    z.x
"""


class ImportExceptionsTest(tests.TestCase):
    def _setUpConfig(self, content):
        tox_ini_path = os.path.join(self.useFixture(fixtures.TempDir()).path,
                                    'tox.ini')

        with open(tox_ini_path, 'w') as tox_ini:
            tox_ini.write(content)

        return config.Config('hacking', tox_ini_path)

    def setUp(self):
        super(ImportExceptionsTest, self).setUp()

    def test_default_import_exceptions(self):
        conf = self._setUpConfig("")
        self.assertEqual(core.DEFAULT_IMPORT_EXCEPTIONS,
                         conf.get_multiple(
                             'import_exceptions',
                             default=core.DEFAULT_IMPORT_EXCEPTIONS))

    def test_import_exceptions(self):
        conf = self._setUpConfig(TEST_TOX_INI)
        self.assertEqual(['a.b.c', 'z.x'],
                         conf.get_multiple('import_exceptions'))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt

from hacking.checks import mock_checks
from hacking import tests


@ddt.ddt
class MockingTestCase(tests.TestCase):
    """This tests hacking checks related to the use of mocking."""

    @ddt.unpack
    @ddt.data(
        (1, 'import mock', None),
        (0, 'from unittest import mock', None),
        (1, 'from mock import patch', None),
        (0, 'from unittest.mock import patch', None),
        (0, 'from mock', '# noqa'))
    def test_H216_hacking_no_third_party_mock(self, err_count, line, noqa):
        if err_count > 0:
            self.assertCheckFails(mock_checks.hacking_no_third_party_mock,
                                  line, noqa)
        else:
            self.assertCheckPasses(mock_checks.hacking_no_third_party_mock,
                                   line, noqa)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import fixtures

from hacking import config
from hacking import tests


TEST_TOX_INI = """[hacking]
option_1 = val_1
option_2 =
    val_2
option_3 =
    val_1,val_2, val_3,
    val_4 , val_5 , val_6,
    val_7
    ,
     val_8
     val_9
"""


class ConfigTest(tests.TestCase):
    def setUp(self):
        tox_ini_path = os.path.join(self.useFixture(fixtures.TempDir()).path,
                                    'tox.ini')

        with open(tox_ini_path, 'w') as tox_ini:
            tox_ini.write(TEST_TOX_INI)
        self.conf = config.Config('hacking', tox_ini_path)
        super(ConfigTest, self).setUp()

    def test_get(self):
        self.assertEqual('val_1', self.conf.get('option_1'))
        self.assertEqual('val_2', self.conf.get('option_2'))
        self.assertEqual('val_3', self.conf.get('option_4', default='val_3'))

    def test_get_multiple(self):
        self.assertEqual(['val_1', 'val_2', 'val_3', 'val_4', 'val_5', 'val_6',
                          'val_7', 'val_8', 'val_9'],
                         self.conf.get_multiple('option_3'))

        self.assertEqual(['val_1', 'val_2'],
                         self.conf.get_multiple('option_4',
                                                default=['val_1', 'val_2']))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import configparser


class Config(object):
    def __init__(self, default_section=None, tox_file='tox.ini'):
        conf = configparser.RawConfigParser()
        conf.read(tox_file)

        self.conf = conf
        self.default_section = default_section

    def get(self, option, section=None, default=None):
        section = section or self.default_section

        if not self.conf.has_section(section):
            return default

        if self.conf.has_option(section, option):
            return self.conf.get(section, option).strip()

        return default

    def get_multiple(self, option, section=None, default=None):
        section = section or self.default_section

        values = self.get(option, section)
        if not values:
            return default

        values = [v.strip() for v in values.split('\n') if v.strip()]
        result = []
        for vals in values:
            result.extend([v.strip() for v in vals.split(',') if v.strip()])

        return result



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ddt

from hacking.checks import other
from hacking import tests


@ddt.ddt
class OthersTestCase(tests.TestCase):
    """This tests hacking checks from the 'other' group."""

    @ddt.unpack
    @ddt.data(
        (1, 'LOG.debug("Test %s" % foo)', None),
        (0, 'LOG.info("Test %s", foo)', None),
        (1, 'LOG.info("Test {}".format(foo))', None),
        (0, 'LOG.error("Test %s" % foo)', '# noqa'),
        (1, 'LOG.debug("Test %s" % "foo")', None),
        (0, 'LOG.debug("Test %s", "foo")', None),
        (0, 'LOG.warning("Test %s", ",".join("%s:%s" % (a, b)))', None),
        (0, "LOG.warning('Testing some stuff')", None))
    def test_H904_hacking_delayed_string_interpolation(
            self, err_count, line, noqa):
        if err_count > 0:
            self.assertCheckFails(other.hacking_delayed_string_interpolation,
                                  line, noqa)
        else:
            self.assertCheckPasses(other.hacking_delayed_string_interpolation,
                                   line, noqa)

    @ddt.unpack
    @ddt.data(
        (False, 'import os\r\nimport sys'),
        (True, 'import os\nimport sys'))
    def test_H903_hacking_no_cr(self, passes, line):
        if passes:
            self.assertCheckPasses(other.hacking_no_cr, line)
        else:
            self.assertCheckFails(other.hacking_no_cr, line)
